/* Rotate the image according to the current scale mode, scaling as needed.
 * 
 * This will read the image from disk if necessary,
 * and it will rotate the image at the appropriate time
 * (when the image is at its smallest).
 *
 * This is the routine that should be called by external callers:
 * callers should never need to call RotateImage.
 *
 * degrees is the increment from the current rotation (curRot).
 */
void ScaleAndRotate(PhoImage* img, int degrees)
{
    int true_width = img->trueWidth;
    int true_height = img->trueHeight;
    int cur_width = img->curWidth;
    int cur_height = img->curHeight;
    int new_width;
    int new_height;
    int aspect_changing;
    int unrot_new_height, unrot_new_width;

    if (gDebug)
        printf("ScaleAndRotate(%d (cur = %d))\n", degrees, img->curRot);

    /* degrees should be between 0 and 360 */
    degrees = (degrees + 360) % 360;

    /* Is the aspect ratio changing? */
    aspect_changing = ((degrees % 180) != 0);

    /* First, load the image if we haven't already, to get true w/h */
    if (true_width == 0 || true_height == 0) {
        if (gDebug) printf("Loading, first time, from ScaleAndRotate!\n");
        LoadImageFromFile(img);
    }

    /* Adjust aspect ratio of true width and height for rotation */
    if (aspect_changing) {
        SWAP(true_width, true_height);
        SWAP(cur_width, cur_height);
    }

    /* Fullsize: display always at real resolution,
     * even if it's too big to fit on the screen.
     */
    if (gScaleMode == PHO_SCALE_FULLSIZE) {
        new_width = true_width;
        new_height = true_height;
    }

    /* Normal: display at full size unless it won't fit the screen,
     * in which case scale it down.
     */
#define NORMAL_SCALE_SLOP 5
    else if (gScaleMode == PHO_SCALE_NORMAL
             || gScaleMode == PHO_SCALE_SCREEN_RATIO) {
        int max_width = gMonitorWidth;
        int max_height = gMonitorHeight;
        int diff;

        new_width = true_width;
        new_height = true_height;

        if (new_width > max_width || new_height > max_height) {
            ScaleToFit(&new_width, &new_height, max_width, max_height);
        }

        new_width *= gScaleRatio;
        new_height *= gScaleRatio;

        /* Now w and h hold the desired sizes.  See if we're close already. */
        diff = abs(cur_width - new_width)
            + abs(cur_height - new_height);
        if (diff < NORMAL_SCALE_SLOP) {
            new_width = cur_width;
            new_height = cur_height;
        }
    }

    else if (gScaleMode == PHO_SCALE_IMG_RATIO) {
        int diff;
        new_width = true_width * gScaleRatio;
        new_height = true_height * gScaleRatio;

        /* See if we're close */
        diff = abs(cur_width - new_width)
               + abs(cur_height - new_height);
        if (diff < NORMAL_SCALE_SLOP) {
            new_width = cur_width;
            new_height = cur_height;
        }
    }
        
    /* Fullscreen: Scale either up or down if necessary to make
     * the largest dimension match the screen size.
     */
#define FULLSCREEN_SCALE_SLOP 20
    else if (gScaleMode == PHO_SCALE_FULLSCREEN) {
        /* If we're in presentation mode, then we need to scale
         * to the current size of the window, not the monitor,
         * because in xinerama gdk_window_fullscreen() will fullscreen
         * onto only one monitor, but gdk_screen_width() gives the
         * width of the full xinerama setup.
         */
        int diffx = abs(cur_width - gMonitorWidth);
        int diffy = abs(cur_height - gMonitorHeight);
        double xratio, yratio;
        gint screenwidth, screenheight;

        if (gDisplayMode == PHO_DISPLAY_PRESENTATION)
            gtk_window_get_size(GTK_WINDOW(gWin), &screenwidth, &screenheight);
        else {
            screenwidth = gMonitorWidth;
            screenheight = gMonitorHeight;
        }

        if (diffx < FULLSCREEN_SCALE_SLOP || diffy < FULLSCREEN_SCALE_SLOP) {
            /* XXX these get overwritten in a few lines! */
            new_width = cur_width;
            new_height = cur_height;
        }

        xratio = (double)screenwidth / true_width;
        yratio = (double)screenheight / true_height;

        /* Use xratio for the more extreme of the two */
        if (xratio > yratio) xratio = yratio;
        new_width = xratio * true_width;
        new_height = xratio * true_height;
    }
    else {
        /* Shouldn't ever happen, means gScaleMode is bogus */
        printf("Internal error: Unknown scale mode %d\n", gScaleMode);
        new_width = cur_width;
        new_height = cur_height;
    }
    /*
     * Finally, we're done with the scaling modes.
     * Time to do the scaling and rotation,
     * reloading the image if needed.
     */

    /* First figure out if we're getting bigger and hence need to reload. */
    if ((new_width > cur_width || new_height > cur_height)
        && (cur_width < true_width && cur_height < true_height)) {
        if (gDebug)
            printf("Getting bigger, from %dx%d to %dx%d -- need to reload\n",
                   cur_width, cur_height, new_width, new_height);

        /* image->curRot is going to be set back to zero when we load
         * from file, so adjust current planned rotation accordingly.
         */
        degrees = (degrees + img->curRot + 360) % 360;
            /* Now it's the absolute end rot desired */

        img->curRot = 0;
        LoadImageFromFile(img);
        printf("After loading from file, pixbuf is %dx%d, curImg %dx%d,\n",
               gdk_pixbuf_get_width(gImage), gdk_pixbuf_get_height(gImage),
               gCurImage->curWidth, gCurImage->curHeight);
        printf("  cur is %dx%d\n", cur_width, cur_height);

        /* Now we no longer care whether aspect is changing from the old rot.
         * Is it changing from 0?
         */
        aspect_changing = (((degrees + 360) % 180) != 0);
        if (aspect_changing) {
            cur_width = true_width = gdk_pixbuf_get_height(gImage);
            cur_height = true_height = gdk_pixbuf_get_width(gImage);
        }
        else {
            cur_width = true_width = gdk_pixbuf_get_width(gImage);
            cur_height = true_height = gdk_pixbuf_get_height(gImage);
        }
        img->trueWidth = true_width;
        img->trueHeight = true_height;
    }

    /* new_* are the sizes we want after rotation. But we need
     * to compare them now to the sizes before rotation. So we
     * need to compensate for rotation.
     */

    /* If we're going to scale up, then do the rotation first,
     * before scaling. Otherwise, scale down first then rotate.
     */
    if (degrees != 0 && (new_width > cur_width || new_height > cur_height)) {
        if (gDebug) printf("Rotating before scaling up\n");
        RotateImage(img, degrees);
        degrees = 0;    /* finished with rotation */
        unrot_new_width = new_width;
        unrot_new_height = new_height;
    }
    else if (aspect_changing) {
        unrot_new_height = new_width;
        unrot_new_width = new_height;
    } else {
        unrot_new_width = new_width;
        unrot_new_height = new_height;
    }

    /* Do the scaling (thought we'd never get there!) */
    if (unrot_new_width != img->curWidth || unrot_new_height != img->curHeight)
    {
        GdkPixbuf* newimage = gdk_pixbuf_scale_simple(gImage,
                                                      unrot_new_width,
                                                      unrot_new_height,
                                                      GDK_INTERP_NEAREST);
        printf("Scale to %dx%d\n", unrot_new_width, unrot_new_height);
        /* scale_simple apparently has no error return; if it fails,
         * it still returns a pixbuf but width and height are -1.
         * Hope this undocumented behavior doesn't change!
         */
        if (!newimage || gdk_pixbuf_get_width(newimage) < 1) {
            printf("\007Error scaling up to %d x %d: probably out of memory\n",
                   unrot_new_width, unrot_new_height);
            if (newimage)
                gdk_pixbuf_unref(newimage);
            Prompt("Couldn't scale up: probably out of memory", "Bummer", 0,
                   "\n ", "");
            return;
        }
        if (gImage)
            gdk_pixbuf_unref(gImage);
        gImage = newimage;

        img->curWidth = gdk_pixbuf_get_width(gImage);
        img->curHeight = gdk_pixbuf_get_height(gImage);
    }

    //img->curWidth = gdk_pixbuf_get_width(gImage);
    //img->curWidth = gdk_pixbuf_get_height(gImage);
    printf("After scaling, pixbuf is %dx%d, curImg %dx%d,\n",
               gdk_pixbuf_get_width(gImage), gdk_pixbuf_get_height(gImage),
               gCurImage->curWidth, gCurImage->curHeight);
    printf("  cur is %dx%d\n", cur_width, cur_height);
    //img->curWidth = cur_width;
    //img->curHeight = cur_height;
    /* If we didn't rotate before, do it now. */
    if (degrees != 0)
        RotateImage(img, degrees);

#ifdef NOTDEF
    printf("**** End of ScaleAndRotate: curRot is %d, cur %dx%d, true %dx%d\n",
           gCurImage->curRot, gCurImage->curWidth, gCurImage->curHeight,
           gCurImage->trueWidth, gCurImage->trueHeight);
#endif

    /* We've finished making our changes. Now we may need to make
     * changes in the window size or position.
     */
    PrepareWindow();
}

